/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	unit test for interpreter								 */
/* --------------------------------------------------------- */

#include "../parser.h"
#include "../interpreter.h"
#include "../parser.h"
#include "../zb_error.h"
#include "../system_resource.h"
#include <iostream>

using namespace std;

#define	INT_DATA( n )	(n) & 0x0FF, ((n) >> 8) & 0x0FF, ((n) >> 16) & 0x0FF, ((n) >> 24) & 0x0FF

namespace {
	int number_of_error = 0;
}

/* --------------------------------------------------------- */
void check( bool f_assert, const char *p_assertion_case ) {

	if( f_assert ) return;
	cout << p_assertion_case << endl;
	number_of_error++;
	return;
}

/* --------------------------------------------------------- */
void test_term1( void ) {

	const unsigned char test[] = {
		c_parser::OP_PLUS,
		c_parser::PC_INUMBER,
		INT_DATA( 0x100 ),
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	ex.expression( v, p );
	int type = v.get()->get_type();
	check( type == c_base_type::TYPE_INTEGER, " +0x100 ̕]ʂ integer ł" );
	check( v.get()->get_integer() == 0x100, " +0x100 ̕]ʂ 0x100 ł" );
}

/* --------------------------------------------------------- */
void test_term2( void ) {

	const unsigned char test[] = {
		c_parser::OP_MINUS,
		c_parser::PC_INUMBER,
		INT_DATA( 0x100 ),
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	ex.expression( v, p );
	int type = v.get()->get_type();
	check( type == c_base_type::TYPE_INTEGER, " -0x100 ̕]ʂ integer ł" );
	check( v.get()->get_integer() == -0x100, " -0x100 ̕]ʂ -0x100 ł" );
}

/* --------------------------------------------------------- */
void test_term3( void ) {

	const unsigned char test[] = {
		c_parser::OP_MINUS,
		c_parser::OP_MINUS,
		c_parser::PC_INUMBER,
		INT_DATA( 0x100 ),
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	ex.expression( v, p );
	int type = v.get()->get_type();
	check( type == c_base_type::TYPE_INTEGER, " --0x100 ̕]ʂ integer ł" );
	check( v.get()->get_integer() == 0x100, " --0x100 ̕]ʂ 0x100 ł" );
}

/* --------------------------------------------------------- */
void test_term4( void ) {

	const unsigned char test[] = {
		c_parser::OP_NOT,
		c_parser::PC_INUMBER,
		INT_DATA( 0x100 ),
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	ex.expression( v, p );
	int type = v.get()->get_type();
	check( type == c_base_type::TYPE_INTEGER, " NOT 0x100 ̕]ʂ integer ł" );
	check( v.get()->get_integer() == ~0x100, " NOT 0x100 ̕]ʂ ~0x100 ł" );
}

/* --------------------------------------------------------- */
void test_addision1( void ) {

	const unsigned char test[] = {
		c_parser::PC_INUMBER,
		INT_DATA( 0x234 ),
		c_parser::OP_PLUS,
		c_parser::PC_INUMBER,
		INT_DATA( 0x345 ),
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	ex.expression( v, p );
	int type = v.get()->get_type();
	check( type == c_base_type::TYPE_INTEGER, " 0x234+0x345 ̕]ʂ integer ł" );
	check( v.get()->get_integer() == 0x234+0x345, " 0x234+0x345 ̕]ʂ 0x234+0x345 ł" );
}

/* --------------------------------------------------------- */
void test_addision2( void ) {

	const unsigned char test[] = {
		c_parser::PC_INUMBER,
		INT_DATA( 0x234 ),
		c_parser::OP_PLUS,
		c_parser::PC_INUMBER,
		INT_DATA( 0x345 ),
		c_parser::OP_PLUS,
		c_parser::PC_INUMBER,
		INT_DATA( 0x456 ),
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	ex.expression( v, p );
	int type = v.get()->get_type();
	check( type == c_base_type::TYPE_INTEGER, " 0x234+0x345+0x456 ̕]ʂ integer ł" );
	check( v.get()->get_integer() == 0x234+0x345+0x456, " 0x234+0x345+0x456 ̕]ʂ 0x234+0x345+0x456 ł" );
}

/* --------------------------------------------------------- */
void test_subtraction1( void ) {

	const unsigned char test[] = {
		c_parser::PC_INUMBER,
		INT_DATA( 0x234 ),
		c_parser::OP_MINUS,
		c_parser::PC_INUMBER,
		INT_DATA( 0x345 ),
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	ex.expression( v, p );
	int type = v.get()->get_type();
	check( type == c_base_type::TYPE_INTEGER, " 0x234-0x345 ̕]ʂ integer ł" );
	check( v.get()->get_integer() == 0x234-0x345, " 0x234-0x345 ̕]ʂ 0x234-0x345 ł" );
}

/* --------------------------------------------------------- */
void test_multiplication1( void ) {

	const unsigned char test[] = {
		c_parser::PC_INUMBER,
		INT_DATA( 0x234 ),
		c_parser::OP_MUL,
		c_parser::PC_INUMBER,
		INT_DATA( 0x345 ),
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	ex.expression( v, p );
	int type = v.get()->get_type();
	check( type == c_base_type::TYPE_INTEGER, " 0x234*0x345 ̕]ʂ integer ł" );
	check( v.get()->get_integer() == 0x234*0x345, " 0x234*0x345 ̕]ʂ 0x234*0x345 ł" );
}

/* --------------------------------------------------------- */
void test_division1( void ) {

	const unsigned char test[] = {
		c_parser::PC_INUMBER,
		INT_DATA( 0x345 ),
		c_parser::OP_DIV,
		c_parser::PC_INUMBER,
		INT_DATA( 0x12 ),
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	ex.expression( v, p );
	int type = v.get()->get_type();
	check( type == c_base_type::TYPE_INTEGER, " 0x345/0x12 ̕]ʂ integer ł" );
	check( v.get()->get_integer() == 0x345/0x12, " 0x345/0x12 ̕]ʂ 0x345/0x12 ł" );
}

/* --------------------------------------------------------- */
void test_division2( void ) {

	const unsigned char test[] = {
		c_parser::PC_INUMBER,
		INT_DATA( 0x345 ),
		c_parser::OP_DIV,
		c_parser::PC_INUMBER,
		INT_DATA( 0 ),
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	int error_code = ex.expression( v, p );
	check( error_code == c_zb_error::ERROR_DIVISION_BY_ZERO, " 0x345/0 ̕]ʂ [ZG[ ł" );
}

/* --------------------------------------------------------- */
void test_remainder1( void ) {

	const unsigned char test[] = {
		c_parser::PC_INUMBER,
		INT_DATA( 0x345 ),
		c_parser::OP_MOD,
		c_parser::PC_INUMBER,
		INT_DATA( 0x12 ),
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	ex.expression( v, p );
	int type = v.get()->get_type();
	check( type == c_base_type::TYPE_INTEGER, " 0x345 MOD 0x12 ̕]ʂ integer ł" );
	check( v.get()->get_integer() == 0x345%0x12, " 0x345 MOD 0x12 ̕]ʂ 0x345%0x12 ł" );
}

/* --------------------------------------------------------- */
void test_remainder2( void ) {

	const unsigned char test[] = {
		c_parser::PC_INUMBER,
		INT_DATA( 0x345 ),
		c_parser::OP_MOD,
		c_parser::PC_INUMBER,
		INT_DATA( 0 ),
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	int error_code = ex.expression( v, p );
	check( error_code == c_zb_error::ERROR_DIVISION_BY_ZERO, " 0x345 MOD 0 ̕]ʂ [ZG[ ł" );
}

/* --------------------------------------------------------- */
void test_and1( void ) {

	const unsigned char test[] = {
		c_parser::PC_INUMBER,
		INT_DATA( 0x345 ),
		c_parser::OP_AND,
		c_parser::PC_INUMBER,
		INT_DATA( 0x123 ),
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	ex.expression( v, p );
	int type = v.get()->get_type();
	check( type == c_base_type::TYPE_INTEGER, " 0x345 AND 0x123 ̕]ʂ integer ł" );
	check( v.get()->get_integer() == (0x345 & 0x123), " 0x345 AND 0x123 ̕]ʂ 0x345 & 0x123 ł" );
}

/* --------------------------------------------------------- */
void test_or1( void ) {

	const unsigned char test[] = {
		c_parser::PC_INUMBER,
		INT_DATA( 0x345 ),
		c_parser::OP_OR,
		c_parser::PC_INUMBER,
		INT_DATA( 0x123 ),
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	ex.expression( v, p );
	int type = v.get()->get_type();
	check( type == c_base_type::TYPE_INTEGER, " 0x345 OR 0x123 ̕]ʂ integer ł" );
	check( v.get()->get_integer() == (0x345 | 0x123), " 0x345 OR 0x123 ̕]ʂ 0x345 | 0x123 ł" );
}

/* --------------------------------------------------------- */
void test_real1( void ) {

	const unsigned char test[] = {
		c_parser::PC_RNUMBER,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF0, 0x3F,		/* 1.0 */
		c_parser::OP_PLUS,
		c_parser::PC_RNUMBER,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x24, 0x40,		/* 10.0 */
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	ex.expression( v, p );
	int type = v.get()->get_type();
	check( type == c_base_type::TYPE_REAL, " 1.0 + 10.0 ̕]ʂ real ł" );
	check( v.get()->get_real() == 1.0+10.0, " 1.0 + 10.0 ̕]ʂ 1.0 + 10.0 ł" );
}

/* --------------------------------------------------------- */
void test_real2( void ) {

	const unsigned char test[] = {
		c_parser::PC_INUMBER,
		INT_DATA( 1 ),
		c_parser::OP_PLUS,
		c_parser::PC_RNUMBER,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x24, 0x40,		/* 10.0 */
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	ex.expression( v, p );
	int type = v.get()->get_type();
	check( type == c_base_type::TYPE_REAL, " 1 + 10.0 ̕]ʂ real ł" );
	check( v.get()->get_real() == 1+10.0, " 1 + 10.0 ̕]ʂ 1 + 10.0 ł" );
}

/* --------------------------------------------------------- */
void test_comparator1( void ) {

	const unsigned char test[] = {
		c_parser::PC_RNUMBER,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF0, 0x3F,		/* 1.0 */
		c_parser::OP_NOT_EQU,
		c_parser::PC_RNUMBER,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x24, 0x40,		/* 10.0 */
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;

	ex.expression( v, p );
	int type = v.get()->get_type();
	check( type == c_base_type::TYPE_INTEGER, " 1.0 <> 10.0 ̕]ʂ integer ł" );
	check( v.get()->get_integer() == -1, " 1.0 <> 10.0 ̕]ʂ -1 ł" );
}

/* --------------------------------------------------------- */
void test_string1( void ) {

	const unsigned char test[] = {
		c_parser::PC_STRING,
		3,
		'A', 'B', 'C',
		c_parser::OP_PLUS,
		c_parser::PC_STRING,
		4,
		'D', 'E', 'F', 'G',
		c_parser::OP_COLON,
	};

	c_expression ex( 0 );
	c_value v;
	const unsigned char *p = test;
	const char *p_str;
	int length;

	ex.expression( v, p );
	int type = v.get()->get_type();
	check( type == c_base_type::TYPE_STRING, " \"ABC\"+ \"DEFG\" ̕]ʂ string ł" );
	v.get()->get_string( p_str, length );
	check( length == 7, " \"ABC\"+ \"DEFG\" ̕]ʂ 7 ł" );
	check( memcmp( p_str, "ABCDEFG", length ) == 0 , " \"ABC\"+ \"DEFG\" ̕]ʂ \"ABCDEFG\" ł" );
}

/* --------------------------------------------------------- */
int main( void ) {

	test_term1();
	test_term2();
	test_term3();
	test_term4();
	test_addision1();
	test_addision2();
	test_subtraction1();
	test_multiplication1();
	test_division1();
	test_division2();
	test_remainder1();
	test_remainder2();
	test_and1();
	test_or1();
	test_real1();
	test_real2();
	test_comparator1();
	test_string1();

	if( number_of_error ) {
		cout << "found " << number_of_error << " error(s)." << endl;
	}
	else {
		cout << "success of all." << endl;
	}
	return 0;
}
