/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	main window and entry point.							 */
/* --------------------------------------------------------- */

#include <qpe/qpeapplication.h>
#include <qdirectpainter_qws.h>
#include <pthread.h>
#include <unistd.h>
#include "qt_basic_e.h"

#include <stdio.h>

/* --------------------------------------------------------- */
namespace {
	c_basic_e *p_basic_e;
	const char *p_cmd_line;

	class c_callback: public c_basic_e_callback {
	public:
		qt_basic_e		*p_window;
		pthread_t		thread;
		pthread_mutex_t	mutex;

		c_callback() {
			pthread_mutex_init( &mutex, NULL );
		}

		~c_callback() {
			pthread_mutex_destroy( &mutex );
		}

		void sleep( unsigned int msec ) {
			usleep( msec * 1000 );
		}

		void lock( void ) {
			pthread_mutex_lock( &mutex );
		}

		void unlock( void ) {
			pthread_mutex_unlock( &mutex );
		}

		void update_screen( void ) {
			QDirectPainter dp( p_window );
			ushort *p = reinterpret_cast<ushort*>( dp.frameBuffer() );
			p_basic_e->draw( p );
		}

		void update_text( void ) {
			QDirectPainter dp( p_window );
			ushort *p = reinterpret_cast<ushort*>( dp.frameBuffer() );
			p_basic_e->update_draw( p );
		}
	};

	c_callback callback_instance;

	/* --------------------------------------------------------- */
	void* basic_e_thread( void* ) {
		p_basic_e->run( p_cmd_line );

		qApp->quit();
		return 0;
	}
};

/* --------------------------------------------------------- */
qt_basic_e::qt_basic_e( QWidget *, const char * ):
		QMainWindow( 0, 0, WStyle_Customize | WStyle_NoBorderEx | WType_TopLevel ) {

	f_press_ctrl = false;
}

/* --------------------------------------------------------- */
qt_basic_e::~qt_basic_e() {

	if( p_basic_e == 0 ) return;

	delete p_basic_e;
}

/* --------------------------------------------------------- */
void qt_basic_e::showEvent ( QShowEvent* ) {

	if( p_basic_e == 0 ) {
		QDirectPainter dp( this );
		p_basic_e = new c_basic_e( dp.lineStep() / 2 - dp.xOffset(), dp.width() - dp.xOffset(), dp.lineStep() );
		if( p_basic_e == 0 ) {
			return;
		}
		p_basic_e->initialize( &callback_instance );
		pthread_create( &(callback_instance.thread), NULL, basic_e_thread, NULL );
	}
}

/* --------------------------------------------------------- */
void qt_basic_e::paintEvent( QPaintEvent* ) {

	if( p_basic_e == 0 ) return;

	QDirectPainter dp( this );
	ushort *p = reinterpret_cast<ushort*>( dp.frameBuffer() );
	p_basic_e->draw( p );
}

/* --------------------------------------------------------- */
void qt_basic_e::keyPressEvent( QKeyEvent* p_key_event ) {

	if( p_basic_e == 0 ) return;

	int key_code = p_key_event->key();
	if( key_code == 0x1021 ) {
		/* CTRL key */
		f_press_ctrl = true;
	}
	else if( (key_code & 0xF0) == 0x10 ) {
		if( key_code >= 0x1012 && key_code <= 0x1015 ) {
			/* It is cursor key.	*/
			key_code = ((key_code + 0x0B) & 0x03) | 0x1C;
			if( (key_code == 0x1C) || (key_code == 0x1F) ) {
				key_code = key_code ^ 0x03;
			}
			p_basic_e->key( key_code );
		}
	}
	else if( f_press_ctrl ) {
		if( key_code == 0x0043 ) {
			/* CTRL+C */
			p_basic_e->force_break();
		}
	}
	else {
		p_basic_e->key( p_key_event->ascii() );
	}
}

/* --------------------------------------------------------- */
void qt_basic_e::keyReleaseEvent( QKeyEvent* ) {

	f_press_ctrl = false;
}

/* --------------------------------------------------------- */
void qt_basic_e::closeEvent( QCloseEvent* ) {
}

/* --------------------------------------------------------- */
void qt_basic_e::set_full_screen_mode( void ) {

	setFixedSize( qApp->desktop()->size() );
	showNormal();
	reparent( 0, WType_TopLevel | WStyle_Customize | WStyle_NoBorderEx, QPoint(0, 0) );
	showFullScreen();
}

/* --------------------------------------------------------- */
void qt_basic_e::resizeEvent( QResizeEvent* ) {
	
	if( size() != qApp->desktop()->size() ) {
		set_full_screen_mode();
	}
}

/* --------------------------------------------------------- */
void qt_basic_e::focusInEvent( QFocusEvent* ) {

	set_full_screen_mode();
}

/* --------------------------------------------------------- */
int main( int argc, char **argv ) {
	QPEApplication app( argc, argv );
	qt_basic_e main_window;

	if( argc >= 2 ) {
		p_cmd_line = argv[1];
	}
	else {
		p_cmd_line = 0;
	}

	callback_instance.p_window = &main_window;

	app.setMainWidget( &main_window );
	app.showMainWidget( &main_window );
	return app.exec();
}
