/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	BASIC-E body											 */
/* --------------------------------------------------------- */

#ifndef __BASIC_E_H__
#define __BASIC_E_H__

#include "system_resource.h"
#include "queue.h"

/* --------------------------------------------------------- */
class c_basic_e_exception {
public:
	typedef enum {
		SYSTEM_ERROR,
	} exception_code;

	exception_code code;

	c_basic_e_exception( exception_code c ): code( c ) {
	}

	const char * what( void ) {
		const char *p_message[] = {
			"no error.",
			"system error.",
		};
		return p_message[ code ];
	}
};

/* --------------------------------------------------------- */
class c_basic_e {
private:
	int width;
	int height;
	int line_step;

	c_graphic_layer		graphic_screen;
	c_text_layer		text_screen;
	c_text_editor		text_editor;
	c_interpreter		interpreter;
	c_variable_manager	variable_manager;

	c_system_resource	resource;

	c_keyboard			keyboard;

public:
	c_basic_e( int screen_width, int screen_height, int screen_line_step ) {
		width		= screen_width;
		height		= screen_height;
		line_step	= screen_line_step;
	}
	
	~c_basic_e() {
	}

	void initialize( c_basic_e_callback *p );

	/* control codes */
	enum {
		CC_BS		= 0x08,		/* back space */
		CC_TAB		= 0x09,
		CC_LF		= 0x0A,		/* line feed */
		CC_HOME		= 0x0B,		/* home position */
		CC_CLS		= 0x0C,		/* clear screen */
		CC_CR		= 0x0D,		/* carriage return */
		CC_INS		= 0x12,		/* insert */
		CC_DEL		= 0x13,		/* delete */
		CC_RIGHT	= 0x1C,
		CC_LEFT		= 0x1D,
		CC_UP		= 0x1E,
		CC_DOWN		= 0x1F,
	};

	void draw( unsigned short *p_vram );
	void update_draw( unsigned short *p_vram );
	void key( int key_code );
	void force_break( void );
	void run( const char *p_line );
	void terminate( void ) {
		resource.f_running = false;
	}
};

#endif
