/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	command													 */
/* --------------------------------------------------------- */

#include "parser.h"
#include "interpreter.h"
#include "system_resource.h"
#include "command.h"
#include "zb_error.h"
#include "debug.h"
#include <cstring>
#include <cstdlib>
#include <climits>

namespace {
	const unsigned char end_mark[] = { c_parser::PC_ENDMARK };

	/* --------------------------------------------------------- */
	void cmd_exit( const unsigned char* &, c_system_resource *p ) {
		p->f_running = false;
	}

	/* --------------------------------------------------------- */
	void cmd_run( const unsigned char* &p_code, c_system_resource *p ) {
		dlog.printf( "  run\n" );
		p->p_interpreter->all_file_close();
		p->p_interpreter->dimension_free();
		p->p_interpreter->list.goto_top();
		p_code = p->p_interpreter->list.get_line();
	}

	/* --------------------------------------------------------- */
	void cmd_print( const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		const char *p_str;
		int length;
		
		p->error_code = expression.expression( value, p_code );
		value.convert_to_string();
		value.get()->get_string( p_str, length );
		p->p_text_editor->put_str( p_str, length );
		if( p_code[0] == c_parser::OP_SEMICOLON ) {
			p_code++;
		}
		else {
			p->p_text_editor->put_str( "\xD\xA" );
		}
		p->p_callback->update_text();
	}

	/* --------------------------------------------------------- */
	void cmd_printh( const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		const char *p_str;
		int length;
		
		p->error_code = expression.expression( value, p_code );
		value.convert_to_string();
		value.get()->get_string( p_str, length );
		p->p_text_editor->put_str( p_str, length );
		if( p_code[0] == c_parser::OP_SEMICOLON ) {
			p_code++;
		}
		else {
			p->p_text_editor->put_str( "\xD\xA" );
		}
	}

	/* --------------------------------------------------------- */
	void cmd_locate( const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		int x, y;

		p->error_code = expression.expression( value, p_code );
		x = value.get()->get_integer();
		if( p->error_code ) return;
		if( p_code[0] != c_parser::OP_COMMA ) {
			p->error_code = c_zb_error::ERROR_SYNTAX;
			return;
		}
		p_code++;
		p->error_code = expression.expression( value, p_code );
		y = value.get()->get_integer();
		p->p_text_editor->locate( x, y );
	}

	/* --------------------------------------------------------- */
	void cmd_cls( const unsigned char* &, c_system_resource *p ) {
		
		p->p_text_editor->clear_screen();
	}

	/* --------------------------------------------------------- */
	void cmd_sleep( const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		int wait;

		p->error_code = expression.expression( value, p_code );
		wait = value.get()->get_integer();
		if( p->error_code ) return;
		p->p_callback->sleep( wait );
	}

	/* --------------------------------------------------------- */
	void cmd_save( const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		const char *p_str;
		int length;
		char filename[ 257 ];

		p->error_code = expression.expression( value, p_code );
		value.convert_to_string();
		value.get()->get_string( p_str, length );

		memcpy( filename, p_str, length );
		filename[ length ] = 0;

		p->p_interpreter->list.save( filename );
	}

	/* --------------------------------------------------------- */
	void cmd_load( const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		const char *p_str;
		int length;
		char filename[ 257 ];

		p->error_code = expression.expression( value, p_code );
		value.convert_to_string();
		value.get()->get_string( p_str, length );

		memcpy( filename, p_str, length );
		filename[ length ] = 0;

		p->p_interpreter->list.clear();
		if( !p->p_interpreter->list.load( filename ) ) {
			p->error_code = c_zb_error::ERROR_DEVICE_IO_ERROR;
		}

		p_code = end_mark;
	}

	/* --------------------------------------------------------- */
	void cmd_list( const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		int start_line_no = INT_MIN, end_line_no = INT_MAX;

		if( ! c_parser::is_command_end( p_code[0] ) ) {
			if( p_code[0] != c_parser::OP_COMMA ) {
				p->error_code = expression.expression( value, p_code );
				start_line_no = value.get()->get_integer();
			}
		}
		if( ! c_parser::is_command_end( p_code[0] ) ) {
			if( p_code[0] != c_parser::OP_COMMA ) {
				p->error_code = c_zb_error::ERROR_ILLEGAL_PARAMETER;
				return;
			}
			p_code++;
			if( ! c_parser::is_command_end( p_code[0] ) ) {
				p->error_code = expression.expression( value, p_code );
				end_line_no = value.get()->get_integer();
			}
			else {
				end_line_no = INT_MAX;
			}
		}
		p->p_interpreter->list.list( start_line_no, end_line_no );
	}

	/* --------------------------------------------------------- */
	void cmd_new( const unsigned char* &p_code, c_system_resource *p ) {

		p->p_interpreter->list.clear();
		p_code = 0;
	}

	/* --------------------------------------------------------- */
	void cmd_randomize( const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		int rand_seed = -1;

		if( ! c_parser::is_command_end( p_code[0] ) ) {
			p->error_code = expression.expression( value, p_code );
			rand_seed = value.get()->get_integer();
		}
		srand( static_cast<unsigned int>( rand_seed ) );
	}

	/* --------------------------------------------------------- */
	void cmd_renum( const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		int start_line_no = 100, step = 10;

		if( ! c_parser::is_command_end( p_code[0] ) ) {
			if( p_code[0] != c_parser::OP_COMMA ) {
				p->error_code = expression.expression( value, p_code );
				start_line_no = value.get()->get_integer();
			}
		}
		if( ! c_parser::is_command_end( p_code[0] ) ) {
			if( p_code[0] != c_parser::OP_COMMA ) {
				p->error_code = c_zb_error::ERROR_ILLEGAL_PARAMETER;
				return;
			}
			p_code++;
			if( ! c_parser::is_command_end( p_code[0] ) ) {
				p->error_code = expression.expression( value, p_code );
				step = value.get()->get_integer();
			}
		}
		if( (start_line_no <= 0) || (step <= 0) ) {
			p->error_code = c_zb_error::ERROR_ILLEGAL_PARAMETER;
			return;
		}
		p->p_interpreter->list.renumber( start_line_no, step );
	}

	/* --------------------------------------------------------- */
	void cmd_dim( const unsigned char* &p_code, c_system_resource *p ) {

		p->p_interpreter->dimension_alloc( p_code );
	}

	/* --------------------------------------------------------- */
	void cmd_update_screen( const unsigned char* &, c_system_resource *p ) {
		
		p->p_callback->update_screen();
	}

	/* --------------------------------------------------------- */
	void cmd_set_pixel( const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		int x, y;
		unsigned short color;

		p->error_code = expression.expression( value, p_code );
		x = value.get()->get_integer();
		if( p->error_code ) return;
		if( p_code[0] != c_parser::OP_COMMA ) {
			p->error_code = c_zb_error::ERROR_SYNTAX;
			return;
		}
		p_code++;
		p->error_code = expression.expression( value, p_code );
		y = value.get()->get_integer();
		if( p->error_code ) return;
		if( p_code[0] != c_parser::OP_COMMA ) {
			p->error_code = c_zb_error::ERROR_SYNTAX;
			return;
		}
		p_code++;
		p->error_code = expression.expression( value, p_code );
		color = static_cast<unsigned short>( value.get()->get_integer() );

		p->p_graphic_screen->set_pixel( x, y, color );
	}

	/* --------------------------------------------------------- */
	void cmd_set_text_color( const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		int color;

		p->error_code = expression.expression( value, p_code );
		color = value.get()->get_integer();
		if( p->error_code ) return;
		p->p_text_editor->set_text_color( color );
	}

	/* --------------------------------------------------------- */
	void cmd_end( const unsigned char* &p_code, c_system_resource * ) {
		
		p_code = 0;
	}

	/* --------------------------------------------------------- */
	void cmd_fropen( const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		const char *p_string;
		int length, descriptor;
		char s_file_name[257];

		expression.expression( value, p_code );
		if( p->error_code ) return;
		descriptor = value.get()->get_integer();
		if( p_code[0] != c_parser::OP_COMMA ) {
			p->error_code = c_zb_error::ERROR_ILLEGAL_PARAMETER;
			return;
		}
		p_code++;
		expression.expression( value, p_code );
		if( p->error_code ) return;
		value.convert_to_string();
		value.get()->get_string( p_string, length );

		memcpy( s_file_name, p_string, length );
		s_file_name[ length ] = 0;
		p->p_interpreter->file_open( descriptor, s_file_name, "rb" );
	}

	/* --------------------------------------------------------- */
	void cmd_fwopen( const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		const char *p_string;
		int length, descriptor;
		char s_file_name[257];

		expression.expression( value, p_code );
		if( p->error_code ) return;
		descriptor = value.get()->get_integer();
		if( p_code[0] != c_parser::OP_COMMA ) {
			p->error_code = c_zb_error::ERROR_ILLEGAL_PARAMETER;
			return;
		}
		p_code++;
		expression.expression( value, p_code );
		if( p->error_code ) return;
		value.convert_to_string();
		value.get()->get_string( p_string, length );

		memcpy( s_file_name, p_string, length );
		s_file_name[ length ] = 0;
		p->p_interpreter->file_open( descriptor, s_file_name, "wb" );
	}

	/* --------------------------------------------------------- */
	void cmd_fclose( const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		int descriptor;

		expression.expression( value, p_code );
		if( p->error_code ) return;
		descriptor = value.get()->get_integer();

		p->p_interpreter->file_close( descriptor );
	}

	/* --------------------------------------------------------- */
	void cmd_fprint( const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		const char *p_string;
		int length, descriptor;
		
		expression.expression( value, p_code );
		if( p->error_code ) return;
		descriptor = value.get()->get_integer();
		if( p_code[0] != c_parser::OP_COMMA ) {
			p->error_code = c_zb_error::ERROR_ILLEGAL_PARAMETER;
			return;
		}
		p_code++;
		expression.expression( value, p_code );
		if( p->error_code ) return;
		value.convert_to_string();
		value.get()->get_string( p_string, length );
		p->p_interpreter->file_print( descriptor, p_string, length );

		if( p_code[0] == c_parser::OP_SEMICOLON ) {
			p_code++;
		}
		else {
			p->p_interpreter->file_print( descriptor, "\xD\xA", 2 );
		}
	}
}

/* --------------------------------------------------------- */
const c_command_list c_command::command_list[] = {
	{	"exit"				,	cmd_exit				},
	{	"run"				,	cmd_run					},
	{	"print"				,	cmd_print				},
	{	"printh"			,	cmd_printh				},
	{	"locate"			,	cmd_locate				},
	{	"cls"				,	cmd_cls					},
	{	"sleep"				,	cmd_sleep				},
	{	"save"				,	cmd_save				},
	{	"load"				,	cmd_load				},
	{	"list"				,	cmd_list				},
	{	"new"				,	cmd_new					},
	{	"randomize"			,	cmd_randomize			},
	{	"renum"				,	cmd_renum				},
	{	"dim"				,	cmd_dim					},
	{	"update_scr"		,	cmd_update_screen		},
	{	"set_pixel"			,	cmd_set_pixel			},
	{	"set_text_color"	,	cmd_set_text_color		},
	{	"end"				,	cmd_end					},
	{	"fropen"			,	cmd_fropen				},
	{	"fwopen"			,	cmd_fwopen				},
	{	"fclose"			,	cmd_fclose				},
	{	"fprint"			,	cmd_fprint				},
};

/* --------------------------------------------------------- */
int c_command::get_code( const char* p_name ) {
	unsigned int i;

	for( i = 0; i < sizeof(command_list)/sizeof(c_command_list); i++ ) {
		if( strcmp( p_name, command_list[i].p_name ) == 0 ) {
			return i;
		}
	}
	return -1;	/* syntax error */
}
