/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	function												 */
/* --------------------------------------------------------- */

#include "system_resource.h"
#include "parser.h"
#include "interpreter.h"
#include "function.h"
#include "zb_error.h"
#include "graphic_layer.h"
#include <cstring>
#include <cstdlib>
#include <cmath>
#include <limits.h>

namespace {

	/* --------------------------------------------------------- */
	void fnc_rnd( c_value& term_value, const unsigned char* &, c_system_resource * ) {

		term_value.change_real( (double)rand() / RAND_MAX );
	}

	/* --------------------------------------------------------- */
	void fnc_sin( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;

		p->error_code = expression.expression( value, p_code );
		term_value.change_real( sin( value.get()->get_real() ) );
	}

	/* --------------------------------------------------------- */
	void fnc_cos( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;

		p->error_code = expression.expression( value, p_code );
		term_value.change_real( cos( value.get()->get_real() ) );
	}

	/* --------------------------------------------------------- */
	void fnc_tan( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;

		p->error_code = expression.expression( value, p_code );
		term_value.change_real( tan( value.get()->get_real() ) );
	}

	/* --------------------------------------------------------- */
	void fnc_log( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;

		p->error_code = expression.expression( value, p_code );
		term_value.change_real( log( value.get()->get_real() ) );
	}

	/* --------------------------------------------------------- */
	void fnc_sqrt( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;

		p->error_code = expression.expression( value, p_code );
		term_value.change_real( sqrt( value.get()->get_real() ) );
	}

	/* --------------------------------------------------------- */
	void fnc_exp( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;

		p->error_code = expression.expression( value, p_code );
		term_value.change_real( exp( value.get()->get_real() ) );
	}

	/* --------------------------------------------------------- */
	void fnc_log10( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;

		p->error_code = expression.expression( value, p_code );
		term_value.change_real( log10( value.get()->get_real() ) );
	}

	/* --------------------------------------------------------- */
	void fnc_abs( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;

		p->error_code = expression.expression( value, p_code );
		term_value.change_real( abs( value.get()->get_real() ) );
	}

	/* --------------------------------------------------------- */
	void fnc_asin( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;

		p->error_code = expression.expression( value, p_code );
		term_value.change_real( asin( value.get()->get_real() ) );
	}

	/* --------------------------------------------------------- */
	void fnc_acos( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;

		p->error_code = expression.expression( value, p_code );
		term_value.change_real( acos( value.get()->get_real() ) );
	}
	/* --------------------------------------------------------- */
	void fnc_atan( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;

		p->error_code = expression.expression( value, p_code );
		term_value.change_real( atan( value.get()->get_real() ) );
	}

	/* --------------------------------------------------------- */
	void fnc_sinh( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;

		p->error_code = expression.expression( value, p_code );
		term_value.change_real( sinh( value.get()->get_real() ) );
	}

	/* --------------------------------------------------------- */
	void fnc_cosh( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;

		p->error_code = expression.expression( value, p_code );
		term_value.change_real( cosh( value.get()->get_real() ) );
	}

	/* --------------------------------------------------------- */
	void fnc_tanh( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;

		p->error_code = expression.expression( value, p_code );
		term_value.change_real( tanh( value.get()->get_real() ) );
	}

	/* --------------------------------------------------------- */
	void fnc_inkey( c_value& term_value, const unsigned char* &, c_system_resource *p ) {
		
		term_value.change_integer( p->p_keyboard->pop_key() );
	}

	/* --------------------------------------------------------- */
	void fnc_rgb( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		int r, g, b;

		p->error_code = expression.expression( value, p_code );
		r = (value.get()->get_integer() ) >> 1;
		if( p_code[0] != c_parser::OP_COMMA ) {
			p->error_code = c_zb_error::ERROR_ILLEGAL_PARAMETER;
			return;
		}
		p_code++;
		p->error_code = expression.expression( value, p_code );
		g = (value.get()->get_integer() );
		if( p_code[0] != c_parser::OP_COMMA ) {
			p->error_code = c_zb_error::ERROR_ILLEGAL_PARAMETER;
			return;
		}
		p_code++;
		p->error_code = expression.expression( value, p_code );
		b = (value.get()->get_integer() ) >> 1;

		if( r <  0 ) r = 0;
		if( g <  0 ) g = 0;
		if( b <  0 ) b = 0;
		if( r > 31 ) r = 31;
		if( g > 63 ) g = 63;
		if( b > 31 ) b = 31;

		term_value.change_integer( rgb_color( r, g, b ) );
	}

	/* --------------------------------------------------------- */
	void fnc_mid( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value svalue, ivalue;
		int start_pos, length, string_length;
		const char *p_string;

		p->error_code = expression.expression( svalue, p_code );
		svalue.convert_to_string();
		if( p_code[0] != c_parser::OP_COMMA ) {
			p->error_code = c_zb_error::ERROR_ILLEGAL_PARAMETER;
			return;
		}
		svalue.get()->get_string( p_string, string_length );
		p_code++;
			
		p->error_code = expression.expression( ivalue, p_code );
		start_pos = ivalue.get()->get_integer();
		if( p_code[0] == c_parser::OP_COMMA ) {
			p_code++;
			p->error_code = expression.expression( ivalue, p_code );
			if( p->error_code ) return;
			length = ivalue.get()->get_integer();
		}
		else {
			length = INT_MAX;
		}
		if( start_pos < 0 ) {
			start_pos = 0;
		}
		if( start_pos >= string_length ) {
			length = 0;
		}
		if( (start_pos + length) > string_length ) {
			length = string_length - start_pos;
		}
		if( length < 0 ) {
			length = 0;
		}
		term_value.change_string( p_string + start_pos, length );
	}

	/* --------------------------------------------------------- */
	void fnc_len( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		int length;
		const char *p_string;

		p->error_code = expression.expression( value, p_code );
		value.convert_to_string();
		value.get()->get_string( p_string, length );
		term_value.change_integer( length );
	}

	/* --------------------------------------------------------- */
	void fnc_finput( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		char s_line[ 257 ];
		int length, descriptor;
		
		expression.expression( value, p_code );
		if( p->error_code ) return;
		descriptor = value.get()->get_integer();
		length = p->p_interpreter->file_input( descriptor, s_line, sizeof(s_line) ); 
		if( p->error_code ) return;
		term_value.change_string( s_line, length );
	}

	/* --------------------------------------------------------- */
	void fnc_feof( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		int descriptor;

		expression.expression( value, p_code );
		if( p->error_code ) return;
		descriptor = value.get()->get_integer();
		term_value.change_integer( p->p_interpreter->file_eof( descriptor ) );
	}

	/* --------------------------------------------------------- */
	void fnc_hex( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		unsigned int num;
		int column;
		char s_hex[ sizeof(int) * 2 ];
		const char hex_char[] = "0123456789ABCDEF"; 

		expression.expression( value, p_code );
		if( p->error_code ) return;
		num = static_cast<unsigned>( value.get()->get_integer() );

		column = sizeof(s_hex) - 1;
		if( num == 0 ) {
			s_hex[ column-- ] = '0';
		}
		else {
			while( num ) {
				s_hex[ column-- ] = hex_char[ num & 0x0F ];
				num >>= 4;
			}
		}
		term_value.change_string( s_hex + column + 1, sizeof(s_hex) - column - 1 );
	}

	/* --------------------------------------------------------- */
	void fnc_bin( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		unsigned int num; 
		int column;
		char s_bin[ sizeof(int) * 8 ];
		const char bin_char[] = "01"; 

		expression.expression( value, p_code );
		if( p->error_code ) return;
		num = static_cast<unsigned>( value.get()->get_integer() );

		column = sizeof(s_bin) - 1;
		if( num == 0 ) {
			s_bin[ column-- ] = '0';
		}
		else {
			while( num ) {
				s_bin[ column-- ] = bin_char[ num & 0x01 ];
				num >>= 1;
			}
		}
		term_value.change_string( s_bin + column + 1, sizeof(s_bin) - column - 1 );
	}

	/* --------------------------------------------------------- */
	void fnc_chr( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		int num;
		char c;

		expression.expression( value, p_code );
		if( p->error_code ) return;
		num = value.get()->get_integer();

		c = static_cast<char>( num & 0xFF );
		term_value.change_string( &c, 1 );
	}

	/* --------------------------------------------------------- */
	void fnc_asc( c_value& term_value, const unsigned char* &p_code, c_system_resource *p ) {
		c_expression expression( p );
		c_value value;
		int length;
		const char *p_string;

		p->error_code = expression.expression( value, p_code );
		value.convert_to_string();
		value.get()->get_string( p_string, length );
		if( length == 0 ) {
			term_value.change_integer( 0 );
		}
		else {
			term_value.change_integer( static_cast<int>( p_string[0] ) );
		}
	}
}

/* --------------------------------------------------------- */
const c_function_list c_function::function_list[] = {
	{	"rnd"		,	fnc_rnd			,	c_parser::PC_RNUMBER	},
	{	"sin"		,	fnc_sin			,	c_parser::PC_RNUMBER	},
	{	"cos"		,	fnc_cos			,	c_parser::PC_RNUMBER	},
	{	"tan"		,	fnc_tan			,	c_parser::PC_RNUMBER	},
	{	"log"		,	fnc_log			,	c_parser::PC_RNUMBER	},
	{	"sqrt"		,	fnc_sqrt		,	c_parser::PC_RNUMBER	},
	{	"exp"		,	fnc_exp			,	c_parser::PC_RNUMBER	},
	{	"log10"		,	fnc_log10		,	c_parser::PC_RNUMBER	},
	{	"abs"		,	fnc_abs			,	c_parser::PC_RNUMBER	},
	{	"asin"		,	fnc_asin		,	c_parser::PC_RNUMBER	},
	{	"acos"		,	fnc_acos		,	c_parser::PC_RNUMBER	},
	{	"atan"		,	fnc_atan		,	c_parser::PC_RNUMBER	},
	{	"sinh"		,	fnc_sinh		,	c_parser::PC_RNUMBER	},
	{	"cosh"		,	fnc_cosh		,	c_parser::PC_RNUMBER	},
	{	"tanh"		,	fnc_tanh		,	c_parser::PC_RNUMBER	},
	{	"inkey"		,	fnc_inkey		,	c_parser::PC_INUMBER	},
	{	"rgb"		,	fnc_rgb			,	c_parser::PC_INUMBER	},
	{	"mid"		,	fnc_mid			,	c_parser::PC_STRING		},
	{	"len"		,	fnc_len			,	c_parser::PC_INUMBER	},
	{	"finput"	,	fnc_finput		,	c_parser::PC_STRING		},
	{	"feof"		,	fnc_feof		,	c_parser::PC_INUMBER	},
	{	"hex"		,	fnc_hex			,	c_parser::PC_STRING		},
	{	"chr"		,	fnc_chr			,	c_parser::PC_STRING		},
	{	"bin"		,	fnc_bin			,	c_parser::PC_STRING		},
};

/* --------------------------------------------------------- */
int c_function::get_code( const char* p_name ) {
	unsigned int i;

	for( i = 0; i < sizeof(function_list)/sizeof(c_function_list); i++ ) {
		if( strcmp( p_name, function_list[i].p_name ) == 0 ) {
			return i;
		}
	}
	return -1;	/* syntax error */
}
