/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	parser													 */
/* --------------------------------------------------------- */

#ifndef __PARSER_H__
#define __PARSER_H__

#include <cctype>

class c_program_line;

class c_parser {
private:
	char line[ 256 ];
	unsigned char precompiled_code[ 256 * 10 ];
	int length;
	int line_no;
	class c_system_resource	*p_res;

	void analyze_number( int& index, const char* &p_string );
	void analyze_symbol( int& index, const char* &p_string );
	void analyze_operator( int& index, const char* &p_string );
	void analyze_string( int& index, const char* &p_string );
	void analyze_label( int& index, const char* &p_string, bool f_set, c_program_line* p_line );
	int get_operator_id( const char * &p_operator );

	bool is_name_char( int c ) {
		return( isalpha(c) || isdigit(c) || (c == '_') );
	}
public:
	c_parser( void );

	enum {
		/* constant value */
		PC_INUMBER	= 0,	/* 0:	integer number */
		PC_RNUMBER,			/* 1:	real number */
		PC_STRING,			/* 2:	string */

		/* variable */
		PC_IVAR,			/* 3:	integer variable */
		PC_RVAR,			/* 4:	real variable */
		PC_SVAR,			/* 5:	string variable */
		PC_IARRAY,			/* 6:	integer number array */
		PC_RARRAY,			/* 7:	real number array */
		PC_SARRAY,			/* 8:	string array */
		PC_LABEL,			/* 9:	label */

		/* subroutine */
		PC_COMMAND,			/* 10:	command */
		PC_FUNCTION,		/* 11:	function */

		/* mark */
		PC_ENDMARK,			/* 12:	line end */

		/* reserve word */
		PC_FOR,				/* 13:	*/
		PC_TO,				/* 14:	*/
		PC_STEP,			/* 15:	*/
		PC_NEXT,			/* 16:	*/
		PC_IF,				/* 17:	*/
		PC_THEN,			/* 18:	*/
		PC_ELSE,			/* 19:	*/
		PC_GOTO,			/* 20:	*/
		PC_GOSUB,			/* 21:	*/
		PC_RETURN,			/* 22:	*/

		/* operator */
		PC_OPERATOR,		/* 23-:	operator */
	};

	static bool is_variable( unsigned char c ) {
		return( (c >= PC_IVAR) && (c <= PC_LABEL) );
	}

	static int get_next_position( const unsigned char *p_code, int index ) {
		switch( p_code[ index ] ) {
		case PC_INUMBER:
		case PC_COMMAND:
		case PC_FUNCTION:
			return index + 1 + sizeof(int);
		case PC_RNUMBER:
			return index + 1 + sizeof(double);
		case PC_STRING:
			return index + 2 + p_code[ index + 1 ];
		default:
			if( p_code[ index ] >= PC_OPERATOR ) {
				return index + 1;
			}
			break;
		}
		return index + 1 + sizeof(void*);
	}

	static void* get_ptr( const unsigned char *p ) {
		int i;
		void *r;
		unsigned char *d = reinterpret_cast<unsigned char*>( &r );

		for( i = 0; i < static_cast<int>(sizeof(void*)); i++ ) {
			d[i] = p[i];
		}
		return r;
	}

	static int put_ptr( unsigned char *r, void* p ) {
		int i;
		unsigned char *d = reinterpret_cast<unsigned char*>( &p );

		for( i = 0; i < static_cast<int>(sizeof(void*)); i++ ) {
			r[i] = d[i];
		}
		return sizeof(void*);
	}

	/* operator code, index of get_operator_id()::p_operator_list[] with offset PC_OPERATOR. */
	enum {
		OP_AND = PC_OPERATOR,
		OP_OR,
		OP_XOR,
		OP_NOT,
		OP_MOD,
		OP_PLUS,
		OP_MINUS,
		OP_MUL,
		OP_DIV,
		OP_L_BLACKET,
		OP_R_BLACKET,
		OP_EQU,
		OP_LT_EQU,
		OP_GT_EQU,
		OP_NOT_EQU,
		OP_LSHIFT,
		OP_RSHIFT,
		OP_LT,
		OP_GT,
		OP_COLON,
		OP_SEMICOLON,
		OP_COMMA,
		OP_POWER
	};

	/* check line No. -1: no operation, 0: direct run, others: line No. */ 
	int check_line_no( const char* &p_string );
	int precompile( const char *p_string, c_program_line* p_line );

	static bool is_command_end( unsigned char c ) {
		return( (c == PC_ENDMARK) || (c == OP_COLON) );
	}

	void set_system_resource( c_system_resource *p ) {
		p_res = p;
	}

	const unsigned char *get_precompiled_code( void ) const {
		return precompiled_code;
	}
	
	void variable_release( void );
	
	int get_length( void ) const {
		return length;
	}

	int get_line_no( void ) const {
		return line_no;
	}

	const char *get_line( void ) const {
		return line;
	}
};

#endif
