/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	program list manager									 */
/* --------------------------------------------------------- */

#include "program_list.h"
#include "system_resource.h"
#include "zb_error.h"
#include <cstdio>

/* --------------------------------------------------------- */
void c_program_list::delete_line( int line_no ) {
	c_program_line *p, *p_next = 0;

	p = p_list;
	if( p ) {
		p_next = p->get_next();
	}
	while( p ) {
		if( p->get_line_no() == line_no ) {
			delete p;
			break;
		}
	}
	if( p == p_list ) {
		p_list = p_next;
	}
}

/* --------------------------------------------------------- */
void c_program_list::add( c_parser *p_parser, const char *p_line ) {
	c_program_line *p, *p_new_line, *p_prev = 0;

	int line_no = p_parser->get_line_no();

	if( isspace( *p_line ) ) p_line++;
	for( p = p_list; p != 0; p = p->get_next() ) {
		if( p->get_line_no() == line_no ) {
			/* update */
			p_res->p_variable_manager->line_release( p->get_code(), p->get_length() );
			if( p_parser->precompile( p_line, p ) != -1 ) {
				p->set( p_parser );
			}
			else {
				/* delete line */
				delete p;
				if( p_list == p ) p_list = 0;
			}
			break;
		}
		else if( p->get_line_no() > line_no ) {
			/* insert */
			p_new_line = new c_program_line();
			if( p_new_line == 0 ) {
				p_res->error_code = c_zb_error::ERROR_OUT_OF_MEMORY;
				return;
			}
			if( p_parser->precompile( p_line, p_new_line ) != -1 ) {
				p_new_line->set( p_parser );
				p->insert_prev( p_new_line );
				if( p_prev == 0 ) {
					p_list = p_new_line;
				}
			}
			else {
				/* delete line */
				delete p_new_line;
			}
			break;
		}
		p_prev = p;
	}
	if( p == 0 ) {
		p_new_line = new c_program_line();
		if( p_new_line == 0 ) {
			p_res->error_code = c_zb_error::ERROR_OUT_OF_MEMORY;
			return;
		}
		if( p_parser->precompile( p_line, p_new_line ) != -1 ) {
			p_new_line->set( p_parser );
			if( p_list == 0 ) {
				/* add top */
				p_list = p_new_line;
			}
			else {
				/* add bottom */
				p_prev->insert_next( p_new_line );
			}
		}
		else {
			/* delete line */
			delete p_new_line;
		}
	}
}

/* --------------------------------------------------------- */
void c_program_list::renumber( int start_line_no, int step ) {
	c_program_line *p;

	for( p = p_list; p != 0; p = p->get_next() ) {
		p->set_line_no( start_line_no );
		start_line_no += step;
	}
}

/* --------------------------------------------------------- */
void c_program_list::clear( void ) {
	c_program_line *p;

	while( p_list ) {
		p = p_list;
		p_list = p_list->get_next();
		delete p;
	}
	goto_top();
}

/* --------------------------------------------------------- */
bool c_program_list::save( const char *p_filename ) {
	c_program_line *p;
	FILE *p_file;
	char *p_line;

	p_file = fopen( p_filename, "wb" );
	if( p_file == 0 ) {
		p_res->error_code = c_zb_error::ERROR_DEVICE_IO_ERROR;
		return false;
	}

	p = this->p_list;
	while( p ) {
		p_line = p->get_line();
		fwrite( p_line, 1, strlen( p_line ), p_file );
		fwrite( "\xD\xA", 1, 2, p_file );
		p = p->get_next();
	}
	fclose( p_file );
	return true;
}

/* --------------------------------------------------------- */
bool c_program_list::load( const char *p_filename, int line_no, int step_no ) {
	char s_line[ 256 ], s_line_no[ 32 ];
	const char *p_line_no;
	c_parser parser;
	c_program_line program_line;
	FILE *p_file;
	int i;

	p_file = fopen( p_filename, "rb" );
	if( p_file == 0 ) {
		p_res->error_code = c_zb_error::ERROR_CANNOT_OPEN_FILE;
		return false;
	}

	parser.set_system_resource( p_res );
	while( !feof( p_file ) ) {
		s_line[0] = 0;
		if( fgets( s_line, sizeof(s_line), p_file ) == 0 ) {
			break;
		}
		for( i = 0; s_line[i]; i++ ) {
			if( (s_line[i] == '\xD') || (s_line[i] == '\xA') ) {
				s_line[i] = 0;
				break;
			}
		}
		sprintf( s_line_no, "%d", line_no );
		p_line_no = s_line_no;
		parser.check_line_no( p_line_no );
		add( &parser, s_line );
		if( p_res->error_code != c_zb_error::NO_PROBLEM ) {
			return false;
		}
		line_no += step_no;
	}

	goto_bottom();
	return true;
}

/* --------------------------------------------------------- */
void c_program_list::list( int start_line_no, int end_line_no ) {
	c_program_line *p;
	char s_line_no[ 32 ];
	int line_no;

	p = p_list;
	while( p ) {
		line_no = p->get_line_no();
		if( (line_no >= start_line_no) && (line_no <= end_line_no) ) {
			sprintf( s_line_no, "%d ", line_no );
			p_res->p_text_editor->put_str( s_line_no );
			p_res->p_text_editor->put_str( p->get_line() );
			p_res->p_text_editor->put_str( "\xD\xA" );
		}
		p = p->get_next();
	}
}
