/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	program list											 */
/* --------------------------------------------------------- */

#ifndef __PROGRAM_LIST_H__
#define __PROGRAM_LIST_H__

#include <cstring>
#include "parser.h"
#include "zb_error.h"

using namespace std;

class c_program_line {
private:
	c_program_line *p_prev;
	c_program_line *p_next;
	int line_no;

	unsigned char *p_memory;
	char *p_line;
	unsigned char *p_precompiled_code;
	int length;

public:
	c_program_line( void ) {
		length = 0;
		line_no = 0;
		p_precompiled_code = 0;
		p_prev = 0;
		p_next = 0;
		p_memory = 0;
		p_line = 0;
	}

	~c_program_line() {
		reject();
		delete[] p_memory;
	}

	unsigned char *get_code( void ) const {
		return p_precompiled_code;
	}
	
	int get_length( void ) const {
		return length;
	}

	char *get_line( void ) const {
		return (char*)p_line;
	}

	int get_line_no( void ) const {
		return line_no;
	}
	
	void set_line_no( int no ) {
		line_no = no;
	}

	c_program_line *get_next( void ) const {
		return p_next;
	}

	void reject( void ) {
		if( p_prev ) {
			p_prev->p_next = p_next;
		}
		if( p_next ) {
			p_next->p_prev = p_prev;
		}
	}

	void set( const c_parser *p_parser ) {

		delete[] p_memory;

		line_no = p_parser->get_line_no();
		size_t length = strlen( p_parser->get_line() );
		p_memory = new unsigned char[ p_parser->get_length() + (length + 1) ];
		p_precompiled_code = p_memory;
		p_line = (char*)p_memory + p_parser->get_length();
		memcpy( p_precompiled_code, p_parser->get_precompiled_code(), p_parser->get_length() );
		strcpy( p_line, p_parser->get_line() );
	}

	void insert_prev( c_program_line *p_new_prev_line ) {

		p_new_prev_line->reject();
		p_new_prev_line->p_next = this;
		p_new_prev_line->p_prev = p_prev;
		if( p_prev ) p_prev->p_next = p_new_prev_line;
		p_prev = p_new_prev_line;
	}

	void insert_next( c_program_line *p_new_next_line ) {

		p_new_next_line->reject();
		p_new_next_line->p_prev = this;
		p_new_next_line->p_next = p_next;
		if( p_next ) p_next->p_prev = p_new_next_line;
		p_next = p_new_next_line;
	}
};

/* --------------------------------------------------------- */
class c_program_list {
private:
	c_program_line *p_list;
	c_program_line *p_current_line;
	class c_system_resource *p_res;

public:
	c_program_list( void ): p_list(0), p_current_line(0) {
	}
	~c_program_list() {
		clear();
	}

	void set_system_resource( c_system_resource *p ) {
		p_res = p;
	}

	void add( c_parser *p_parser, const char *p_line );
	void clear( void );
	bool is_null( void ) const {
		return( p_list == 0 );
	}
	bool save( const char *p_filename );
	bool load( const char *p_filename, int line_no = 100, int step_no = 10 );
	void list( int start_line_no, int end_line_no );
	void renumber( int start_line_no, int step );

	void delete_line( int line_no );

	void set_current_line_no( int line_no );
	int get_current_line_no( void ) {
		if( p_current_line == 0 ) return 0;
		return p_current_line->get_line_no();
	}
	unsigned char *get_line( void ) {
		if( p_current_line == 0 ) return 0;
		return p_current_line->get_code();
	}
	c_program_line *get_current_line( void ) {
		return p_current_line;
	}
	void goto_ptr( c_program_line* p ) {
		p_current_line = p;
	}
	void goto_top( void ) {
		p_current_line = p_list;
	}
	void goto_bottom( void ) {
		p_current_line = 0;
	}
	void goto_next_line( void ) {
		if( p_current_line ) {
			p_current_line = p_current_line->get_next();
		}
	}
	bool is_bottom( void ) {
		return( p_current_line == 0 );
	}
};

#endif
