/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	text editor												 */
/* --------------------------------------------------------- */

#include "text_editor.h"
#include "basic_e.h"
#include "zb_error.h"
#include "debug.h"
#include <cstdio>

using namespace std;

/* --------------------------------------------------------- */
void c_text_editor::key( int key_code ) {

	put_char( key_code );
	p_res->p_callback->update_text();
}

/* --------------------------------------------------------- */
void c_text_editor::back_space( void ) {

	if( x == 0 ) {
		if( y == 0 ) {
			delete_char();
			return;
		}
		if( (p_res->p_text_screen->get_char( -1, y ) & 0xFF) == 0 ) {
			delete_char();
			return;
		}
	}
	left();
	delete_char();
}

/* --------------------------------------------------------- */
void c_text_editor::tab_indent( void ) {

	
}

/* --------------------------------------------------------- */
void c_text_editor::line_feed( void ) {

	hide_cursor();
	x = 0;
	show_cursor();
}

/* --------------------------------------------------------- */
void c_text_editor::home_position( void ) {

	hide_cursor();
	x = 0;
	y = 0;
	show_cursor();
}

/* --------------------------------------------------------- */
void c_text_editor::locate( int x, int y ) {

	if( x < 0 ) {
		x = 0;
	}
	else if( x >= p_res->p_text_screen->get_columns() ) {
		x = p_res->p_text_screen->get_columns() - 1;
	}
	if( y < 0 ) {
		y = 0;
	}
	else if( y >= p_res->p_text_screen->get_rows() ) {
		y = p_res->p_text_screen->get_rows() - 1;
	}
	this->x = x;
	this->y = y;
}

/* --------------------------------------------------------- */
void c_text_editor::clear_screen( void ) {

	p_res->p_text_screen->clear();
	x = 0;
	y = 0;
	show_cursor();
}

/* --------------------------------------------------------- */
void c_text_editor::put_error( int error_code, int line_no ) {
	char message[ 256 ];
	c_zb_error error;

	if( line_no ) {
		sprintf( message, "\xD\xA%s at %d.\xD\xAOk\xD\xA", error.get_error_message( error_code ), line_no );
		dlog.printf( "  Program stop: %s at %d\n", error.get_error_message( error_code ), line_no );
	}
	else {
		sprintf( message, "\xD\xA%s.\xD\xAOk\xD\xA", error.get_error_message( error_code ) );
		dlog.printf( "  Program stop: %s\n", error.get_error_message( error_code ) );
	}
	set_edit_mode( false );
	put_str( message );
	set_edit_mode( true );
}

/* --------------------------------------------------------- */
void c_text_editor::carriage_return( void ) {
	c_parser parser;
	c_program_line direct_line;

	hide_cursor();
	if( f_edit_mode ) {
		char line_buffer[ 256 ];
		const char *p_line;

		if( !line_detect( line_buffer, sizeof(line_buffer) ) ) {
			put_error( c_zb_error::ERROR_TOO_LONG_LINE, 0 );
			return;
		}
		p_line = line_buffer;
		parser.set_system_resource( p_res );
		int line_no = parser.check_line_no( p_line );
		if( p_res->error_code != c_zb_error::NO_PROBLEM ) {
			put_error( p_res->error_code, line_no );
			return;
		}
		if( line_no == 0 ) {
			/* direct run */
			parser.precompile( p_line, &direct_line );
			direct_line.set( &parser );
			set_edit_mode( false );
			enable_cursor( false );
			x = 0;
			y++;
			if( y >= p_res->p_text_screen->get_rows() ) {
				y--;
				p_res->p_text_screen->scroll_up();
			}
			line_no = p_res->p_interpreter->run( &parser );
			if( p_res->error_code != c_zb_error::NO_PROBLEM ) {
				y--;
				put_error( p_res->error_code, line_no );
				y--;
			}
			else {
				if( x != 0 ) put_str( "\xD\xA" );
				put_str( "Ok" );
			}
			enable_cursor( true );
			set_edit_mode( true );
			y = y + 1;
		}
		else if( line_no == -1 ) {
			/* no operation */
			y = y + 1;
		}
		else {
			/* regist line */
			p_res->p_interpreter->list.add( &parser, p_line );
			y = y + (static_cast<int>(strlen( line_buffer )) + p_res->p_text_screen->get_columns() - 1 ) 
					/ p_res->p_text_screen->get_columns();
		}
		/* with line feed when edit mode. */
		x = 0;
	}
	else {
		y = y + 1;
	}
	if( y >= p_res->p_text_screen->get_rows() ) {
		y--;
		p_res->p_text_screen->scroll_up();
	}
	show_cursor();
}

/* --------------------------------------------------------- */
void c_text_editor::toggle_insert_mode( void ) {

	f_insert_mode = ! f_insert_mode;
}

/* --------------------------------------------------------- */
void c_text_editor::right( void ) {

	hide_cursor();
	x++;
	if( x == p_res->p_text_screen->get_columns() ) {
		y++;
		if( y == p_res->p_text_screen->get_rows() ) {
			y--;
			x--;
		}
		else {
			x = 0;
		}
	}
	show_cursor();
}

/* --------------------------------------------------------- */
void c_text_editor::left( void ) {

	hide_cursor();
	if( x == 0 ) {
		if( y == 0 ) {
			show_cursor();
			return;
		}
		x = p_res->p_text_screen->get_columns() - 1;
		y--;
	}
	else {
		x--;
	}
	show_cursor();
}

/* --------------------------------------------------------- */
void c_text_editor::up( void ) {

	if( y == 0 ) return;
	hide_cursor();
	y--;
	show_cursor();
}

/* --------------------------------------------------------- */
void c_text_editor::down( void ) {

	if( y == (p_res->p_text_screen->get_rows() - 1) ) return;
	hide_cursor();
	y++;
	show_cursor();
}

/* --------------------------------------------------------- */
void c_text_editor::delete_char( void ) {

	p_res->p_text_screen->delete_char( x, y );
	show_cursor();
}

/* --------------------------------------------------------- */
bool c_text_editor::line_detect( char *p_line, int size ) {
	int index;
	unsigned int c;

	/* search line top index */
	index = p_res->p_text_screen->get_columns() * y;
	c = 0;
	while( index && ((p_res->p_text_screen->get_char( index - 1 ) & 0x0FF) != 0) ) {
		index--;
		c++;
		if( int(c) >= size ) return false;
	}

	/* get line */
	while( size && ((c = p_res->p_text_screen->get_char( index )) & 0x0FF) ) {
		*p_line = char(c & 0xFF);
		p_line++;
		index++;
		size--;
	}
	y = index / p_res->p_text_screen->get_columns();
	if( size == 0 ) {
		return false;
	}
	*p_line = 0;
	return true;
}

/* --------------------------------------------------------- */
void c_text_editor::put_char( int char_code ) {

	if( (char_code & 0xE0) == 0 ) {
		/* It is control code */
		switch( char_code ) {
		case c_basic_e::CC_BS		:		/* back space */
			back_space();
			break;
		case c_basic_e::CC_TAB		:
			tab_indent();
			break;
		case c_basic_e::CC_LF		:		/* line feed */
			line_feed();
			break;
		case c_basic_e::CC_HOME		:		/* home position */
			home_position();
			break;
		case c_basic_e::CC_CLS		:		/* clear screen */
			clear_screen();
			break;
		case c_basic_e::CC_CR		:		/* carriage return */
			carriage_return();
			break;
		case c_basic_e::CC_INS		:		/* insert */
			toggle_insert_mode();
			break;
		case c_basic_e::CC_RIGHT	:
			right();
			break;
		case c_basic_e::CC_LEFT		:
			left();
			break;
		case c_basic_e::CC_UP		:
			up();
			break;
		case c_basic_e::CC_DOWN		:
			down();
			break;
		case c_basic_e::CC_DEL		:		/* delete */
			delete_char();
			break;
		}
	}
	else {
		if( f_insert_mode && f_edit_mode ) {
			hide_cursor();
			if( p_res->p_text_screen->insert( x, y, color | (char_code & 0x0FF) ) ) {
				y--;
			}
		}
		else {
			p_res->p_text_screen->put_char( x, y, color | (char_code & 0x0FF) );
		}
		x = x + 1;
		if( x >= p_res->p_text_screen->get_columns() ) {
			x = 0;
			y = y + 1;
			if( y >= p_res->p_text_screen->get_rows() ) {
				y--;
				p_res->p_text_screen->scroll_up();
			}
		}
		show_cursor();
	}
}

/* --------------------------------------------------------- */
void c_text_editor::put_str( const char *p_string ) {

	while( *p_string ) {
		put_char( *p_string );
		p_string++;
	}
}

/* --------------------------------------------------------- */
void c_text_editor::put_str( const char *p_string, int length ) {

	while( length ) {
		put_char( *p_string );
		p_string++;
		length--;
	}
}

/* --------------------------------------------------------- */
void c_text_editor::hide_cursor( void ) {
	if( f_visible_cursor ) {
		p_res->p_text_screen->put_char( x, y, p_res->p_text_screen->get_char( x, y ) & ~0x100, false );
	}
}

/* --------------------------------------------------------- */
void c_text_editor::show_cursor( void ) {
	if( f_visible_cursor ) {
		p_res->p_text_screen->put_char( x, y, p_res->p_text_screen->get_char( x, y ) | 0x100, false );
	}
}
