/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	text editor												 */
/* --------------------------------------------------------- */

#ifndef __TEXT_EDITOR_H__
#define __TEXT_EDITOR_H__

#include "text_layer.h"

using namespace std;

class c_text_editor {
private:
	class c_system_resource *p_res;
	int x;							/* current cursor position */
	int y;
	int color;						/* current font color */
	bool f_edit_mode;				/* true: edit mode, false: run mode */
	bool f_visible_cursor;			/* true: visible cursor, false: invisible cursor */
	bool f_insert_mode;
	bool line_detect( char *p_line, int size );
public:

	c_text_editor( void ): 
			x(0), y(0), color(0xFFFF0000),
			f_edit_mode(true), f_visible_cursor(true), f_insert_mode(true) {
		p_res = 0;
	}
	
	~c_text_editor() {
	}

	void set_system_resource( c_system_resource *p ) {
		p_res = p;
	}

	void set_edit_mode( bool f_enable ) {
		f_edit_mode = f_enable;
	}

	void set_text_color( int c ) {
		color = (c & 0x0FFFF) << 16;
	}

	void key( int key_code );
	void put_char( int char_code );
	void put_str( const char *p_string );
	void put_str( const char *p_string, int length );
	void locate( int x, int y );

	void back_space( void );
	void tab_indent( void );
	void line_feed( void );
	void home_position( void );
	void clear_screen( void );
	void carriage_return( void );
	void toggle_insert_mode( void );
	void right( void );
	void left( void );
	void up( void );
	void down( void );
	void delete_char( void );
	void put_error( int error_code, int line_no );

	void hide_cursor( void );
	void show_cursor( void );
	void enable_cursor( bool f_enable ) {
		f_visible_cursor = f_enable;
	}
};

#endif
