/* --------------------------------------------------------- */
/*	BASIC-E TEXT layer										 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	text screen layer										 */
/* --------------------------------------------------------- */

#ifndef __TEXT_LAYER_H__
#define __TEXT_LAYER_H__

#include <cstring>
#include "graphic_layer.h"

class c_text_layer {
private:
	int columns;
	int rows;
	int width;
	int height;
	int line_step;

	/* (MSB) {rrrrrggggggbbbbbXXXXXXURXccccccc} (LSB)	*/
	/*	rrrrr ... color (red element)					*/
	/*	gggggg .. color (green element)					*/
	/*	bbbbb ... color (blue element)					*/
	/*	X ....... don't care							*/
	/*	U ....... update flag							*/
	/*	R ....... reverse (cursor) flag					*/
	/*	ccccccc . character code						*/
	unsigned int *p_text_vram;
	c_graphic_layer *p_graphic;
	
public:
	c_text_layer( void ): columns(0), rows(0), p_text_vram(0) {
	}
	
	~c_text_layer() {
		release();
	}

	void set_graphic( c_graphic_layer *p ) {
		p_graphic = p;
	}

	void create( int width, int height, int step ) {
		release();
		this->width		= width;
		this->height	= height;
		line_step		= step / 2;
		columns			= height / 8;
		rows			= width  / 16;
		p_text_vram = new unsigned int [ columns * rows + 1 ];	/* "+1" is guardian for delete_char(). */
		clear();
	}

	void clear( void ) {
		int x, size;

		size = columns * rows + 1;
		for( x = 0; x < size; x++ ) {
			p_text_vram[ x ] = 0xFFFF0200;
		}
	}

	void release( void ) {
		delete[] p_text_vram;
		p_text_vram = 0;
	}

	int get_rows( void ) const {
		return rows;
	}

	int get_columns( void ) const {
		return columns;
	}

	unsigned int get_char( int column, int row ) const {
		return p_text_vram[ column + row * columns ];
	}
	
	unsigned int get_char( int index ) const {
		return p_text_vram[ index ];
	}

	void fill_space( int address, int column ) {

		for( int i = 1; i <= column; i++ ) {
			if( (p_text_vram[ address - i ] & 0xFF) != 0 ) {
				break;
			}
			p_text_vram[ address - i ] |= 0x220;
		}
	}
	
	void put_char( int column, int row, unsigned int c, bool f_fill_space = true ) {
		int address;

		address = column + row * columns;
		if( address && f_fill_space ) {
			fill_space( address, column );
		}
		p_text_vram[ column + row * columns ] = c | 0x200;
	}

	bool insert( int column, int row, unsigned int c );

	void draw( unsigned short *p_vram );
	void update( unsigned short *p_vram );

	void scroll_up( void );
	void delete_char( int column, int row );
};

#endif
