/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	variant value											 */
/* --------------------------------------------------------- */

#include "variant_value.h"
#include "interpreter.h"
#include "text_editor.h"
#include "zb_error.h"
#include <cstdio>
#include <cstring>
#include <cmath>

using namespace std;

/* --------------------------------------------------------- */
c_value::c_value( void ) {

	p = dynamic_cast<c_base_type*>( &unknown_type );
}

/* --------------------------------------------------------- */
void c_value::change_integer( int value ) {

	integer_type.set( value );
	p = &integer_type;
}

/* --------------------------------------------------------- */
void c_value::change_real( double value ) {

	real_type.set( value );
	p = &real_type;
}

/* --------------------------------------------------------- */
void c_value::change_string( const char *p_string, int length ) {

	string_type.set( p_string, length );
	p = &string_type;
}

/* --------------------------------------------------------- */
void c_value::convert_to_string( void ) {
	char buffer[ 256 ];
	int length;

	if( p->get_type() == c_base_type::TYPE_STRING ) return;
	p->convert_to_string( buffer, length );
	string_type.set( buffer, length );
	p = &string_type;
}

/* --------------------------------------------------------- */
int c_base_type::get_integer( void ) {
	return 0;
}

/* --------------------------------------------------------- */
double c_base_type::get_real( void ) {
	return 0.;
}

/* --------------------------------------------------------- */
void c_base_type::get_string( const char* &p, int &length ) {
	p = "";
	length = 0;
}

/* ========================================================= */
void c_base_type::convert_to_string( char *, int &length ) {
	length = 0;
}

/* --------------------------------------------------------- */
int c_base_type::minus( void ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::plus( void ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::not( void ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::addition( c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::subtraction( c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::lshift( c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::rshift( c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::multiplication( c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::division( c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::power( c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::remainder( c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::and( c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::or( c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::xor( c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::equ( int &, c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::not_equ( int &, c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::gt( int &, c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::gt_equ( int &, c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::lt( int &, c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* --------------------------------------------------------- */
int c_base_type::lt_equ( int &, c_base_type* ) {

	return c_zb_error::ERROR_SYNTAX;
}

/* ========================================================= */
int c_integer_type::get_integer( void ) {
	return value;
}

/* --------------------------------------------------------- */
double c_integer_type::get_real( void ) {
	return double( value );
}

/* --------------------------------------------------------- */
void c_integer_type::convert_to_string( char* p, int &length ) {

	sprintf( p, "%d", value );
	length = static_cast<int>( strlen( p ) );
}

/* --------------------------------------------------------- */
int c_integer_type::minus( void ) {

	value = -value;
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::plus( void ) {

	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::not( void ) {

	value = ~value;
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::addition( c_base_type* p_operand ) {

	value += p_operand->get_integer();
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::subtraction( c_base_type* p_operand ) {

	value -= p_operand->get_integer();
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::lshift( c_base_type* p_operand ) {
	int sh;

	sh = p_operand->get_integer();
	if( sh >= 32 ) {
		value = 0;
	}
	else {
		value <<= sh;
	}
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::rshift( c_base_type* p_operand ) {
	int sh;

	sh = p_operand->get_integer();
	if( sh >= 32 ) {
		value >>= 31;
	}
	else {
		value >>= sh;
	}
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::multiplication( c_base_type* p_operand ) {

	value *= p_operand->get_integer();
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::division( c_base_type* p_operand ) {

	int n = p_operand->get_integer();
	if( n == 0 ) {
		return c_zb_error::ERROR_DIVISION_BY_ZERO;
	}
	value /= n;
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::remainder( c_base_type* p_operand ) {

	int n = p_operand->get_integer();
	if( n == 0 ) {
		return c_zb_error::ERROR_DIVISION_BY_ZERO;
	}
	value %= n;
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::power( c_base_type* p_operand ) {
	int i, n, v;

	v = value;
	n = p_operand->get_integer();
	if( n > 0 ) {
		for( i = 1; i < n; i++ ) {
			value *= v;
		}
	}
	else if( n < 0 ) {
		value = 0;
	}
	else {
		value = 1;
	}
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::and( c_base_type* p_operand ) {

	value &= p_operand->get_integer();
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::or( c_base_type* p_operand ) {

	value |= p_operand->get_integer();
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::xor( c_base_type* p_operand ) {

	value ^= p_operand->get_integer();
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::equ( int &result, c_base_type* p_operand ) {

	result = -( value == p_operand->get_integer());
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::not_equ( int &result, c_base_type* p_operand ) {

	result = -( value != p_operand->get_integer());
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::gt( int &result, c_base_type* p_operand ) {

	result = -( value > p_operand->get_integer());
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::gt_equ( int &result, c_base_type* p_operand ) {

	result = -( value >= p_operand->get_integer());
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::lt( int &result, c_base_type* p_operand ) {

	result = -( value < p_operand->get_integer());
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_integer_type::lt_equ( int &result, c_base_type* p_operand ) {

	result = -(value <= p_operand->get_integer());
	return c_zb_error::NO_PROBLEM;
}

/* ========================================================= */
int c_real_type::get_integer( void ) {
	return int( value );
}

/* --------------------------------------------------------- */
double c_real_type::get_real( void ) {
	return value;
}

/* --------------------------------------------------------- */
void c_real_type::convert_to_string( char* p, int &length ) {

	sprintf( p, "%f", value );
	length = static_cast<int>( strlen( p ) );
}

/* --------------------------------------------------------- */
int c_real_type::minus( void ) {

	value = -value;
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_real_type::plus( void ) {

	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_real_type::addition( c_base_type* p_operand ) {

	value += p_operand->get_real();
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_real_type::subtraction( c_base_type* p_operand ) {

	value -= p_operand->get_real();
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_real_type::multiplication( c_base_type* p_operand ) {

	value *= p_operand->get_real();
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_real_type::division( c_base_type* p_operand ) {
	double r;

	r = p_operand->get_real();
	if( r == 0 ) {
		return c_zb_error::ERROR_DIVISION_BY_ZERO;
	}
	value /= r;
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_real_type::remainder( c_base_type* ) {

	return c_zb_error::ERROR_TYPE_MISMATCH;
}

/* --------------------------------------------------------- */
int c_real_type::power( c_base_type* p_operand ) {
	double n;

	n = p_operand->get_real();
	value = pow( value, n );
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_real_type::equ( int &result, c_base_type* p_operand ) {

	result = -( value == p_operand->get_real());
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_real_type::not_equ( int &result, c_base_type* p_operand ) {

	result = -( value != p_operand->get_real());
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_real_type::gt( int &result, c_base_type* p_operand ) {

	result = -( value > p_operand->get_real());
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_real_type::gt_equ( int &result, c_base_type* p_operand ) {

	result = -( value >= p_operand->get_real());
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_real_type::lt( int &result, c_base_type* p_operand ) {

	result = -( value < p_operand->get_real());
	return c_zb_error::NO_PROBLEM;
}

/* --------------------------------------------------------- */
int c_real_type::lt_equ( int &result, c_base_type* p_operand ) {

	result = -(value <= p_operand->get_real());
	return c_zb_error::NO_PROBLEM;
}

/* ========================================================= */
void c_string_type::get_string( const char* &p, int &length ) {

	p		= this->value;
	length	= this->length;
}

/* --------------------------------------------------------- */
void c_string_type::convert_to_string( char* p, int &length ) {

	if( this->length ) memcpy( p, value, this->length );
	length = this->length;
}

/* --------------------------------------------------------- */
int c_string_type::addition( c_base_type* p_operand ) {
	char buffer[ 256 ];
	int new_length, operand_length;
	const char *p;

	if( p_operand->get_type() == c_base_type::TYPE_STRING ) {
		p_operand->get_string( p, operand_length );
	}
	else if( p_operand->get_type() == c_base_type::TYPE_REAL ) {
		sprintf( buffer, "%f", p_operand->get_real() );
		operand_length = static_cast<int>( strlen( buffer ) );
		p = buffer;
	}
	else {
		sprintf( buffer, "%d", p_operand->get_integer() );
		operand_length = static_cast<int>( strlen( buffer ) );
		p = buffer;
	}
	new_length = length + operand_length;
	if( new_length > static_cast<int>(sizeof(value)) ) {
		return c_zb_error::ERROR_OUT_OF_STRING_SPACE;
	}
	memcpy( value + length, p, operand_length );
	length = new_length;
	return c_zb_error::NO_PROBLEM;
}
