/* --------------------------------------------------------- */
/*	BASIC-E													 */
/*	Copyright(C)2006 t.hara									 */
/* ========================================================= */
/*	variant value											 */
/* --------------------------------------------------------- */

#ifndef __VARIANT_VALUE_H__
#define __VARIANT_VALUE_H__

#include <cstring>

class c_base_type {
protected:
	class c_expression		*p_expression;
	unsigned char			*p_code;
	int						type;

public:
	enum {
		TYPE_UNKNOWN = 0,
		TYPE_INTEGER,
		TYPE_REAL,
		TYPE_STRING,
	};

	int get_type( void ) const {
		return type;
	}

	virtual int get_integer( void );
	virtual double get_real( void );
	virtual void get_string( const char* &p, int &length );
	virtual void convert_to_string( char* p, int &length );

	virtual int minus( void );
	virtual int plus( void );
	virtual int not( void );
	virtual int addition( c_base_type* p_operand );
	virtual int subtraction( c_base_type* p_operand );
	virtual int lshift( c_base_type* p_operand );
	virtual int rshift( c_base_type* p_operand );
	virtual int multiplication( c_base_type* p_operand );
	virtual int division( c_base_type* p_operand );
	virtual int power( c_base_type* p_operand );
	virtual int remainder( c_base_type* p_operand );
	virtual int and( c_base_type* p_operand );
	virtual int or( c_base_type* p_operand );
	virtual int xor( c_base_type* p_operand );

	virtual int equ( int &result, c_base_type* p_operand );
	virtual int not_equ( int &result, c_base_type* p_operand );
	virtual int gt( int &result, c_base_type* p_operand );
	virtual int gt_equ( int &result, c_base_type* p_operand );
	virtual int lt( int &result, c_base_type* p_operand );
	virtual int lt_equ( int &result, c_base_type* p_operand );
};

/* --------------------------------------------------------- */
class c_unknown_type: public c_base_type {
public:
	c_unknown_type( void ) {
		type = TYPE_UNKNOWN;
	}
};

/* --------------------------------------------------------- */
class c_integer_type: public c_base_type {
protected:
	int value;

public:
	c_integer_type( void ) {
		type = TYPE_INTEGER;
	}

	int get_integer( void );
	double get_real( void );
	void convert_to_string( char* p, int &length );

	int minus( void );
	int plus( void );
	int not( void );
	int addition( c_base_type* p_operand );
	int subtraction( c_base_type* p_operand );
	int lshift( c_base_type* p_operand );
	int rshift( c_base_type* p_operand );
	int multiplication( c_base_type* p_operand );
	int division( c_base_type* p_operand );
	int remainder( c_base_type* p_operand );
	int power( c_base_type* p_operand );
	int and( c_base_type* p_operand );
	int or( c_base_type* p_operand );
	int xor( c_base_type* p_operand );

	int equ( int &result, c_base_type* p_operand );
	int not_equ( int &result, c_base_type* p_operand );
	int gt( int &result, c_base_type* p_operand );
	int gt_equ( int &result, c_base_type* p_operand );
	int lt( int &result, c_base_type* p_operand );
	int lt_equ( int &result, c_base_type* p_operand );

public:
	void set( int v ) {
		value = v;
	}
};

/* --------------------------------------------------------- */
class c_real_type: public c_base_type {
protected:
	double value;

public:
	c_real_type( void ) {
		type = TYPE_REAL;
	}

	int get_integer( void );
	double get_real( void );
	void convert_to_string( char* p, int &length );

	int minus( void );
	int plus( void );
	int addition( c_base_type* p_operand );
	int subtraction( c_base_type* p_operand );
	int multiplication( c_base_type* p_operand );
	int division( c_base_type* p_operand );
	int remainder( c_base_type* p_operand );
	int power( c_base_type* p_operand );

	int equ( int &result, c_base_type* p_operand );
	int not_equ( int &result, c_base_type* p_operand );
	int gt( int &result, c_base_type* p_operand );
	int gt_equ( int &result, c_base_type* p_operand );
	int lt( int &result, c_base_type* p_operand );
	int lt_equ( int &result, c_base_type* p_operand );

public:
	void set( double v ) {
		value = v;
	}
};

/* --------------------------------------------------------- */
class c_string_type: public c_base_type {
protected:
	char value[ 256 ];
	int length;

public:
	c_string_type( void ) {
		type = TYPE_STRING;
	}

	void get_string( const char* &p, int &length );
	void convert_to_string( char* p, int &length );

	int addition( c_base_type* p_operand );

public:
	void set( const char *p_string, int len ) {
		std::memcpy( value, p_string, len );
		length = len;
	}
};

/* --------------------------------------------------------- */
class c_value {
private:
	c_base_type		*p;
	c_unknown_type	unknown_type;
	c_integer_type	integer_type;
	c_real_type		real_type;
	c_string_type	string_type;

public:
	c_value( void );

	void change_integer( int value );
	void change_real( double v );
	void change_string( const char *p_string, int len );
	void convert_to_string( void );

	c_base_type *get( void ) {
		return p;
	}
};

#endif
